# Copyright 2026 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import pytest

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class MissedQtUpdateTestCase(TestCase):
    @pytest.mark.xfail()
    def testMissedQtUpdate(self):
        """
        Testcase where Portage was unable to upgrade from
        Qt 6.9.3 -> Qt 6.10.1 without an explicit redundant
        argument on the command line (bug #968228).

        This was fixed by the "earlier slot operator backtracking"
        patch related to bug #964705.
        """
        ebuilds = {
            "dev-qt/qtbase-6.9.3": {
                "EAPI": "8",
                "SLOT": "6",
                "IUSE": "+nls",
                "RDEPEND": """
                    !<dev-qt/qtconnectivity-6.9.3:6
                    !<dev-qt/qtsvg-6.9.3:6
                    !<dev-qt/qttools-6.9.3:6
                """,
                "PDEPEND": "nls? ( ~dev-qt/qttranslations-6.9.3:6 )",
            },
            "dev-qt/qtbase-6.10.1": {
                "EAPI": "8",
                "SLOT": "6",
                "IUSE": "+nls",
                "RDEPEND": """
                    !<dev-qt/qtconnectivity-6.10.1:6
                    !<dev-qt/qtsvg-6.10.1:6
                    !<dev-qt/qttools-6.10.1:6
                """,
                "PDEPEND": "nls? ( ~dev-qt/qttranslations-6.10.1:6 )",
            },
            "dev-qt/qtconnectivity-6.9.3": {
                "EAPI": "8",
                "SLOT": "6",
                "DEPEND": "~dev-qt/qtbase-6.9.3:6",
            },
            "dev-qt/qtconnectivity-6.10.1": {
                "EAPI": "8",
                "SLOT": "6",
                "DEPEND": "~dev-qt/qtbase-6.10.1:6",
            },
            "dev-qt/qttranslations-6.9.3": {
                "EAPI": "8",
                "SLOT": "6",
                "DEPEND": "~dev-qt/qtbase-6.9.3:6",
                "BDEPEND": "~dev-qt/qttools-6.9.3:6",
            },
            "dev-qt/qttranslations-6.10.1": {
                "EAPI": "8",
                "SLOT": "6",
                "DEPEND": "~dev-qt/qtbase-6.10.1:6",
                "BDEPEND": "~dev-qt/qttools-6.10.1:6",
            },
            "dev-qt/qtsvg-6.9.3": {
                "EAPI": "8",
                "SLOT": "6",
                "DEPEND": "~dev-qt/qtbase-6.9.3:6",
                "RDEPEND": """
                    ~dev-qt/qtbase-6.9.3:6
                """,
            },
            "dev-qt/qtsvg-6.10.1": {
                "EAPI": "8",
                "SLOT": "6",
                "DEPEND": "~dev-qt/qtbase-6.10.1:6",
                "RDEPEND": """
                    ~dev-qt/qtbase-6.10.1:6
                """,
            },
            "dev-qt/qttools-6.9.3": {
                "EAPI": "8",
                "SLOT": "6",
                "DEPEND": "~dev-qt/qtbase-6.9.3:6",
                "RDEPEND": "~dev-qt/qtbase-6.9.3:6",
            },
            "dev-qt/qttools-6.10.1": {
                "EAPI": "8",
                "SLOT": "6",
                "DEPEND": "~dev-qt/qtbase-6.10.1:6",
                "RDEPEND": """
                    ~dev-qt/qtbase-6.10.1:6
                """,
            },
            "dev-python/pyside-6.9.3": {
                "EAPI": "8",
                "SLOT": "6/6.9.3",
                "DEPEND": """
                    =dev-qt/qtbase-6.9.3*:6
                    =dev-qt/qttools-6.9.3*:6
                    =dev-qt/qtconnectivity-6.9.3*:6
                """,
                "RDEPEND": """
                    =dev-qt/qtbase-6.9.3*:6
                    =dev-qt/qttools-6.9.3*:6
                    =dev-qt/qtconnectivity-6.9.3*:6
                """,
            },
            "dev-python/pyside-6.10.1": {
                "EAPI": "8",
                "SLOT": "6/6.10.1",
                "DEPEND": """
                    =dev-qt/qtbase-6.10.1*:6
                    =dev-qt/qtconnectivity-6.10.1*:6
                """,
                "RDEPEND": """
                    =dev-qt/qtbase-6.10.1*:6
                    =dev-qt/qtconnectivity-6.10.1*:6
                """,
            },
            "media-gfx/freecad-1.0.1-r2": {
                "EAPI": "8",
                "SLOT": "6",
                "DEPEND": """
                    dev-qt/qtbase:6
                    dev-qt/qtsvg:6
                    dev-python/pyside:6=
                """,
                "RDEPEND": """
                    dev-qt/qtbase:6
                    dev-qt/qtsvg:6
                    dev-python/pyside:6=
                """,
            },
        }
        installed = {
            "dev-qt/qtbase-6.9.3": {
                "EAPI": "8",
                "SLOT": "6",
                "IUSE": "+nls",
                "USE": "nls",
                "PDEPEND": "nls? ( ~dev-qt/qttranslations-6.9.3:6 )",
            },
            "dev-qt/qtconnectivity-6.9.3": {
                "EAPI": "8",
                "SLOT": "6",
                "DEPEND": "~dev-qt/qtbase-6.9.3:6",
            },
            "dev-qt/qttranslations-6.9.3": {
                "EAPI": "8",
                "SLOT": "6",
                "DEPEND": "~dev-qt/qtbase-6.9.3:6",
                "BDEPEND": "~dev-qt/qttools-6.9.3:6",
            },
            "dev-qt/qtsvg-6.9.3": {
                "EAPI": "8",
                "SLOT": "6",
                "DEPEND": "~dev-qt/qtbase-6.9.3:6",
                "RDEPEND": "~dev-qt/qtbase-6.9.3:6",
            },
            "dev-qt/qttools-6.9.3": {
                "EAPI": "8",
                "SLOT": "6",
                "DEPEND": "~dev-qt/qtbase-6.9.3:6",
                "RDEPEND": "~dev-qt/qtbase-6.9.3:6",
            },
            "dev-python/pyside-6.9.3": {
                "EAPI": "8",
                "SLOT": "6/6.9.3",
                "DEPEND": """
                    =dev-qt/qtbase-6.9.3*:6
                    =dev-qt/qtconnectivity-6.9.3*:6
                """,
                "RDEPEND": """
                    =dev-qt/qtbase-6.9.3*:6
                    =dev-qt/qtconnectivity-6.9.3*:6
                """,
            },
            "media-gfx/freecad-1.0.1-r2": {
                "EAPI": "8",
                "SLOT": "6",
                "DEPEND": """
                    dev-qt/qtbase:6
                    dev-qt/qtsvg:6
                    dev-python/pyside:6/6.9.3=
                """,
                "RDEPEND": """
                    dev-qt/qtbase:6
                    dev-qt/qtsvg:6
                    dev-python/pyside:6/6.9.3=
                """,
            },
        }

        world = ("media-gfx/freecad",)

        test_cases = (
            # The extra pyside atom is sufficient to nudge Portage
            # towards a solution but shouldn't be necessary.
            # ResolverPlaygroundTestCase(
            #     ["@world", "=dev-python/pyside-6.10.1"],
            #     success=True,
            #     options={"--update": True, "--deep": True},
            #     mergelist=[
            #         "dev-qt/qtbase-6.10.1",
            #         "dev-qt/qttools-6.10.1",
            #         "!<dev-qt/qttools-6.10.1:6",
            #         "dev-qt/qttranslations-6.10.1",
            #         "dev-qt/qtconnectivity-6.10.1",
            #         "!<dev-qt/qtconnectivity-6.10.1:6",
            #         "dev-qt/qtsvg-6.10.1",
            #         "!<dev-qt/qtsvg-6.10.1:6",
            #         "dev-python/pyside-6.10.1",
            #         "media-gfx/freecad-1.0.1-r2",
            #     ],
            # ),
            # It should resolve identically (or at least with a solution)
            # without explicit dev-python/pyside, as it's a dependency of
            # media-gfx/freecad.
            ResolverPlaygroundTestCase(
                ["@world"],
                success=True,
                options={"--update": True, "--deep": True},
                mergelist=[
                    "dev-qt/qtbase-6.10.1",
                    "dev-qt/qttools-6.10.1",
                    "!<dev-qt/qttools-6.10.1:6",
                    "dev-qt/qttranslations-6.10.1",
                    "dev-qt/qtconnectivity-6.10.1",
                    "!<dev-qt/qtconnectivity-6.10.1:6",
                    "dev-qt/qtsvg-6.10.1",
                    "!<dev-qt/qtsvg-6.10.1:6",
                    "dev-python/pyside-6.10.1",
                    "media-gfx/freecad-1.0.1-r2",
                ],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 2011-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import multiprocessing
import sys
import tempfile
import traceback

import portage
from portage import os
from portage import shutil
from portage.exception import TryAgain
from portage.tests import TestCase


class LockNonblockTestCase(TestCase):
    def _testLockNonblock(self, env=None):
        tempdir = tempfile.mkdtemp()
        try:
            path = os.path.join(tempdir, "lock_me")
            lock1 = portage.locks.lockfile(path)
            proc = multiprocessing.Process(
                target=self._lock_subprocess, args=(path, env)
            )
            proc.start()
            self.assertEqual(proc.pid > 0, True)
            proc.join()
            self.assertEqual(proc.exitcode, os.EX_OK)

            portage.locks.unlockfile(lock1)
        finally:
            shutil.rmtree(tempdir)

    @staticmethod
    def _lock_subprocess(path, env):
        if env is not None:
            # This serves to implement __PORTAGE_TEST_HARDLINK_LOCKS
            # environment variable inheritance for the multiprocessing
            # forkserver start method.
            os.environ.update(env)
        portage.locks._close_fds()
        # Disable close_fds since we don't exec
        # (see _setup_pipes docstring).
        portage.process._setup_pipes({0: 0, 1: 1, 2: 2}, close_fds=False)
        rval = 2
        try:
            try:
                lock2 = portage.locks.lockfile(path, flags=os.O_NONBLOCK)
            except portage.exception.TryAgain:
                rval = os.EX_OK
            else:
                rval = 1
                portage.locks.unlockfile(lock2)
        except Exception:
            traceback.print_exc()
        sys.exit(rval)

    def testLockNonblock(self):
        self._testLockNonblock()

    def testLockNonblockHardlink(self):
        prev_state = os.environ.pop("__PORTAGE_TEST_HARDLINK_LOCKS", None)
        os.environ["__PORTAGE_TEST_HARDLINK_LOCKS"] = "1"
        try:
            self._testLockNonblock(env={"__PORTAGE_TEST_HARDLINK_LOCKS": "1"})
        finally:
            os.environ.pop("__PORTAGE_TEST_HARDLINK_LOCKS", None)
            if prev_state is not None:
                os.environ["__PORTAGE_TEST_HARDLINK_LOCKS"] = prev_state

    def test_competition_with_same_process(self):
        """
        Test that at attempt to lock the same file multiple times in the
        same process will behave as intended (bug 714480).
        """
        tempdir = tempfile.mkdtemp()
        try:
            path = os.path.join(tempdir, "lock_me")
            lock = portage.locks.lockfile(path)
            self.assertRaises(
                TryAgain, portage.locks.lockfile, path, flags=os.O_NONBLOCK
            )
            portage.locks.unlockfile(lock)
        finally:
            shutil.rmtree(tempdir)



# Copyright 2010-2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ("UseManager",)

from _emerge.Package import Package
from portage import os
from portage.dep import (
    dep_getrepo,
    dep_getslot,
    ExtendedAtomDict,
    remove_slot,
    _get_useflag_re,
)
from portage.eapi import (
    eapi_supports_use_stable,
    eapi_supports_stable_use_forcing_and_masking,
)
from portage.localization import _
from portage.repository.config import allow_profile_repo_deps
from portage.util import (
    grabfile,
    grabdict_package,
    read_corresponding_eapi_file,
    stack_lists,
    writemsg,
)
from portage.versions import _pkg_str

from portage.package.ebuild._config.helper import ordered_by_atom_specificity


class UseManager:
    def __init__(
        self, repositories, profiles, abs_user_config, is_stable, user_config=True
    ):
        # 	file				variable
        # --------------------------------
        # 	repositories
        # --------------------------------
        # 	use.mask			_repo_usemask_dict
        # 	use.stable			_repo_use_stable_dict
        # 	use.stable.mask			_repo_usestablemask_dict
        # 	use.force			_repo_useforce_dict
        # 	use.stable.force		_repo_usestableforce_dict
        # 	package.use.mask		_repo_pusemask_dict
        # 	package.use.stable		_repo_puse_stable_dict
        # 	package.use.stable.mask		_repo_pusestablemask_dict
        # 	package.use.force		_repo_puseforce_dict
        # 	package.use.stable.force	_repo_pusestableforce_dict
        # --------------------------------
        # 	profiles
        # --------------------------------
        # 	use.mask			_usemask_list
        # 	use.stable			_use_stable_list
        # 	use.stable.mask			_usestablemask_list
        # 	use.force			_useforce_list
        # 	use.stable.force		_usestableforce_list
        # 	package.use.mask		_pusemask_list
        # 	package.use.stable		_puse_stable_list
        # 	package.use.stable.mask		_pusestablemask_list
        # 	package.use			_pkgprofileuse
        # 	package.use.force		_puseforce_list
        # 	package.use.stable.force	_pusestableforce_list
        # --------------------------------
        # 	user config
        # --------------------------------
        # 	package.use			_pusedict

        # Dynamic variables tracked by the config class
        # --------------------------------
        # 	profiles
        # --------------------------------
        # 	usemask
        # 	useforce
        # --------------------------------
        # 	user config
        # --------------------------------
        # 	puse

        self._user_config = user_config
        self._is_stable = is_stable
        self._repo_usemask_dict = self._parse_repository_files_to_dict_of_tuples(
            "use.mask", repositories
        )
        self._repo_use_stable_dict = self._parse_repository_files_to_dict_of_tuples(
            "use.stable",
            repositories,
            eapi_filter=eapi_supports_use_stable,
        )
        self._repo_usestablemask_dict = self._parse_repository_files_to_dict_of_tuples(
            "use.stable.mask",
            repositories,
            eapi_filter=eapi_supports_stable_use_forcing_and_masking,
        )
        self._repo_useforce_dict = self._parse_repository_files_to_dict_of_tuples(
            "use.force", repositories
        )
        self._repo_usestableforce_dict = self._parse_repository_files_to_dict_of_tuples(
            "use.stable.force",
            repositories,
            eapi_filter=eapi_supports_stable_use_forcing_and_masking,
        )
        self._repo_pusemask_dict = self._parse_repository_files_to_dict_of_dicts(
            "package.use.mask", repositories
        )
        self._repo_puse_stable_dict = self._parse_repository_files_to_dict_of_dicts(
            "package.use.stable",
            repositories,
            eapi_filter=eapi_supports_use_stable,
        )
        self._repo_pusestablemask_dict = self._parse_repository_files_to_dict_of_dicts(
            "package.use.stable.mask",
            repositories,
            eapi_filter=eapi_supports_stable_use_forcing_and_masking,
        )
        self._repo_puseforce_dict = self._parse_repository_files_to_dict_of_dicts(
            "package.use.force", repositories
        )
        self._repo_pusestableforce_dict = self._parse_repository_files_to_dict_of_dicts(
            "package.use.stable.force",
            repositories,
            eapi_filter=eapi_supports_stable_use_forcing_and_masking,
        )
        self._repo_puse_dict = self._parse_repository_files_to_dict_of_dicts(
            "package.use", repositories
        )

        self._usemask_list = self._parse_profile_files_to_tuple_of_tuples(
            "use.mask", profiles
        )
        self._use_stable_list = self._parse_profile_files_to_tuple_of_tuples(
            "use.stable",
            profiles,
            eapi_filter=eapi_supports_use_stable,
        )
        self._usestablemask_list = self._parse_profile_files_to_tuple_of_tuples(
            "use.stable.mask",
            profiles,
            eapi_filter=eapi_supports_stable_use_forcing_and_masking,
        )
        self._useforce_list = self._parse_profile_files_to_tuple_of_tuples(
            "use.force", profiles
        )
        self._usestableforce_list = self._parse_profile_files_to_tuple_of_tuples(
            "use.stable.force",
            profiles,
            eapi_filter=eapi_supports_stable_use_forcing_and_masking,
        )
        self._pusemask_list = self._parse_profile_files_to_tuple_of_dicts(
            "package.use.mask", profiles
        )
        self._puse_stable_list = self._parse_profile_files_to_tuple_of_dicts(
            "package.use.stable",
            profiles,
            eapi_filter=eapi_supports_use_stable,
            juststrings=True,
        )
        self._pusestablemask_list = self._parse_profile_files_to_tuple_of_dicts(
            "package.use.stable.mask",
            profiles,
            eapi_filter=eapi_supports_stable_use_forcing_and_masking,
        )
        self._pkgprofileuse = self._parse_profile_files_to_tuple_of_dicts(
            "package.use", profiles, juststrings=True
        )
        self._puseforce_list = self._parse_profile_files_to_tuple_of_dicts(
            "package.use.force", profiles
        )
        self._pusestableforce_list = self._parse_profile_files_to_tuple_of_dicts(
            "package.use.stable.force",
            profiles,
            eapi_filter=eapi_supports_stable_use_forcing_and_masking,
        )

        self._pusedict = self._parse_user_files_to_extatomdict(
            "package.use", abs_user_config, user_config
        )

        self.repositories = repositories

    def _parse_file_to_tuple(
        self, file_name, recursive=True, eapi_filter=None, eapi=None, eapi_default="0"
    ):
        """
        @param file_name: input file name
        @type file_name: str
        @param recursive: triggers recursion if the input file is a
                directory
        @type recursive: bool
        @param eapi_filter: a function that accepts a single eapi
                argument, and returns true if the current file type
                is supported by the given EAPI
        @type eapi_filter: callable
        @param eapi: the EAPI of the current profile node, which allows
                a call to read_corresponding_eapi_file to be skipped
        @type eapi: str
        @param eapi_default: the default EAPI which applies if the
                current profile node does not define a local EAPI
        @type eapi_default: str
        @rtype: tuple
        @return: collection of USE flags
        """
        ret = []
        lines = grabfile(file_name, recursive=recursive)
        if eapi is None:
            eapi = read_corresponding_eapi_file(file_name, default=eapi_default)
        if eapi_filter is not None and not eapi_filter(eapi):
            if lines:
                writemsg(
                    _("--- EAPI '%s' does not support '%s': '%s'\n")
                    % (eapi, os.path.basename(file_name), file_name),
                    noiselevel=-1,
                )
            return ()
        useflag_re = _get_useflag_re(eapi)
        for prefixed_useflag in lines:
            if prefixed_useflag[:1] == "-":
                useflag = prefixed_useflag[1:]
            else:
                useflag = prefixed_useflag
            if useflag_re.match(useflag) is None:
                writemsg(
                    _("--- Invalid USE flag in '%s': '%s'\n")
                    % (file_name, prefixed_useflag),
                    noiselevel=-1,
                )
            else:
                ret.append(prefixed_useflag)
        return tuple(ret)

    def _parse_file_to_dict(
        self,
        file_name,
        juststrings=False,
        recursive=True,
        eapi_filter=None,
        user_config=False,
        eapi=None,
        eapi_default="0",
        allow_repo=False,
        allow_build_id=False,
    ):
        """
        @param file_name: input file name
        @type file_name: str
        @param juststrings: store dict values as space-delimited strings
                instead of tuples
        @type juststrings: bool
        @param recursive: triggers recursion if the input file is a
                directory
        @type recursive: bool
        @param eapi_filter: a function that accepts a single eapi
                argument, and returns true if the current file type
                is supported by the given EAPI
        @type eapi_filter: callable
        @param user_config: current file is part of the local
                configuration (not repository content)
        @type user_config: bool
        @param eapi: the EAPI of the current profile node, which allows
                a call to read_corresponding_eapi_file to be skipped
        @type eapi: str
        @param eapi_default: the default EAPI which applies if the
                current profile node does not define a local EAPI
        @type eapi_default: str
        @param allow_build_id: allow atoms to specify a particular
                build-id
        @type allow_build_id: bool
        @rtype: tuple
        @return: collection of USE flags
        """
        ret = {}
        location_dict = {}
        if eapi is None:
            eapi = read_corresponding_eapi_file(file_name, default=eapi_default)
        extended_syntax = eapi is None and user_config
        if extended_syntax:
            ret = ExtendedAtomDict(dict)
        else:
            ret = {}
        allow_repo = allow_repo or extended_syntax
        file_dict = grabdict_package(
            file_name,
            recursive=recursive,
            allow_wildcard=extended_syntax,
            allow_repo=allow_repo,
            verify_eapi=(not extended_syntax),
            eapi=eapi,
            eapi_default=eapi_default,
            allow_build_id=allow_build_id,
            allow_use=False,
        )
        if eapi is not None and eapi_filter is not None and not eapi_filter(eapi):
            if file_dict:
                writemsg(
                    _("--- EAPI '%s' does not support '%s': '%s'\n")
                    % (eapi, os.path.basename(file_name), file_name),
                    noiselevel=-1,
                )
            return ret
        useflag_re = _get_useflag_re(eapi)
        for k, v in file_dict.items():
            useflags = []
            use_expand_prefix = ""
            for prefixed_useflag in v:
                if extended_syntax and prefixed_useflag == "\n":
                    use_expand_prefix = ""
                    continue
                if extended_syntax and prefixed_useflag[-1] == ":":
                    use_expand_prefix = prefixed_useflag[:-1].lower() + "_"
                    continue

                if prefixed_useflag[:1] == "-":
                    useflag = use_expand_prefix + prefixed_useflag[1:]
                    prefixed_useflag = "-" + useflag
                else:
                    useflag = use_expand_prefix + prefixed_useflag
                    prefixed_useflag = useflag
                if useflag_re.match(useflag) is None:
                    writemsg(
                        _("--- Invalid USE flag for '%s' in '%s': '%s'\n")
                        % (k, file_name, prefixed_useflag),
                        noiselevel=-1,
                    )
                else:
                    useflags.append(prefixed_useflag)
            location_dict.setdefault(k, []).extend(useflags)
        for k, v in location_dict.items():
            if juststrings:
                v = " ".join(v)
            else:
                v = tuple(v)
            ret.setdefault(k.cp, {})[k] = v
        return ret

    def _parse_user_files_to_extatomdict(self, file_name, location, user_config):
        ret = ExtendedAtomDict(dict)
        if user_config:
            pusedict = grabdict_package(
                os.path.join(location, file_name),
                recursive=1,
                newlines=1,
                allow_wildcard=True,
                allow_repo=True,
                verify_eapi=False,
                allow_build_id=True,
                allow_use=False,
            )
            for k, v in pusedict.items():
                l = []
                use_expand_prefix = ""
                for flag in v:
                    if flag == "\n":
                        use_expand_prefix = ""
                        continue
                    if flag[-1] == ":":
                        use_expand_prefix = flag[:-1].lower() + "_"
                        continue
                    if flag[0] == "-":
                        nv = "-" + use_expand_prefix + flag[1:]
                    else:
                        nv = use_expand_prefix + flag
                    l.append(nv)
                ret.setdefault(k.cp, {})[k] = tuple(l)

        return ret

    def _parse_repository_files_to_dict_of_tuples(
        self, file_name, repositories, eapi_filter=None
    ):
        ret = {}
        for repo in repositories.repos_with_profiles():
            ret[repo.name] = self._parse_file_to_tuple(
                os.path.join(repo.location, "profiles", file_name),
                eapi_filter=eapi_filter,
                eapi_default=repo.eapi,
            )
        return ret

    def _parse_repository_files_to_dict_of_dicts(
        self, file_name, repositories, eapi_filter=None
    ):
        ret = {}
        for repo in repositories.repos_with_profiles():
            ret[repo.name] = self._parse_file_to_dict(
                os.path.join(repo.location, "profiles", file_name),
                eapi_filter=eapi_filter,
                eapi_default=repo.eapi,
                allow_repo=allow_profile_repo_deps(repo),
                allow_build_id=("build-id" in repo.profile_formats),
            )
        return ret

    def _parse_profile_files_to_tuple_of_tuples(
        self, file_name, locations, eapi_filter=None
    ):
        return tuple(
            self._parse_file_to_tuple(
                os.path.join(profile.location, file_name),
                recursive=profile.portage1_directories,
                eapi_filter=eapi_filter,
                eapi=profile.eapi,
                eapi_default=None,
            )
            for profile in locations
        )

    def _parse_profile_files_to_tuple_of_dicts(
        self, file_name, locations, juststrings=False, eapi_filter=None
    ):
        return tuple(
            self._parse_file_to_dict(
                os.path.join(profile.location, file_name),
                juststrings,
                recursive=profile.portage1_directories,
                eapi_filter=eapi_filter,
                user_config=profile.user_config,
                eapi=profile.eapi,
                eapi_default=None,
                allow_build_id=profile.allow_build_id,
                allow_repo=allow_profile_repo_deps(profile),
            )
            for profile in locations
        )

    def _isStable(self, pkg):
        if self._user_config:
            try:
                return pkg.stable
            except AttributeError:
                # KEYWORDS is unavailable (prior to "depend" phase)
                return False

        try:
            pkg._metadata
        except AttributeError:
            # KEYWORDS is unavailable (prior to "depend" phase)
            return False

        # Since repoman uses different config instances for
        # different profiles, we have to be careful to do the
        # stable check against the correct profile here.
        return self._is_stable(pkg)

    def getUseMask(self, pkg=None, stable=None):
        if pkg is None:
            return frozenset(stack_lists(self._usemask_list, incremental=True))

        slot = None
        cp = getattr(pkg, "cp", None)
        if cp is None:
            slot = dep_getslot(pkg)
            repo = dep_getrepo(pkg)
            pkg = _pkg_str(remove_slot(pkg), slot=slot, repo=repo)
            cp = pkg.cp

        if stable is None:
            stable = self._isStable(pkg)

        usemask = []

        if hasattr(pkg, "repo") and pkg.repo != Package.UNKNOWN_REPO:
            repos = []
            try:
                repos.extend(repo.name for repo in self.repositories[pkg.repo].masters)
            except KeyError:
                pass
            repos.append(pkg.repo)
            for repo in repos:
                usemask.append(self._repo_usemask_dict.get(repo, {}))
                if stable:
                    usemask.append(self._repo_usestablemask_dict.get(repo, {}))
                cpdict = self._repo_pusemask_dict.get(repo, {}).get(cp)
                if cpdict:
                    pkg_usemask = ordered_by_atom_specificity(cpdict, pkg)
                    if pkg_usemask:
                        usemask.extend(pkg_usemask)
                if stable:
                    cpdict = self._repo_pusestablemask_dict.get(repo, {}).get(cp)
                    if cpdict:
                        pkg_usemask = ordered_by_atom_specificity(cpdict, pkg)
                        if pkg_usemask:
                            usemask.extend(pkg_usemask)

        for i, pusemask_dict in enumerate(self._pusemask_list):
            if self._usemask_list[i]:
                usemask.append(self._usemask_list[i])
            if stable and self._usestablemask_list[i]:
                usemask.append(self._usestablemask_list[i])
            cpdict = pusemask_dict.get(cp)
            if cpdict:
                pkg_usemask = ordered_by_atom_specificity(cpdict, pkg)
                if pkg_usemask:
                    usemask.extend(pkg_usemask)
            if stable:
                cpdict = self._pusestablemask_list[i].get(cp)
                if cpdict:
                    pkg_usemask = ordered_by_atom_specificity(cpdict, pkg)
                    if pkg_usemask:
                        usemask.extend(pkg_usemask)

        return frozenset(stack_lists(usemask, incremental=True))

    def getUseForce(self, pkg=None, stable=None):
        if pkg is None:
            return frozenset(stack_lists(self._useforce_list, incremental=True))

        cp = getattr(pkg, "cp", None)
        if cp is None:
            slot = dep_getslot(pkg)
            repo = dep_getrepo(pkg)
            pkg = _pkg_str(remove_slot(pkg), slot=slot, repo=repo)
            cp = pkg.cp

        if stable is None:
            stable = self._isStable(pkg)

        useforce = []

        if hasattr(pkg, "repo") and pkg.repo != Package.UNKNOWN_REPO:
            repos = []
            try:
                repos.extend(repo.name for repo in self.repositories[pkg.repo].masters)
            except KeyError:
                pass
            repos.append(pkg.repo)
            for repo in repos:
                useforce.append(self._repo_useforce_dict.get(repo, {}))
                if stable:
                    useforce.append(self._repo_usestableforce_dict.get(repo, {}))
                cpdict = self._repo_puseforce_dict.get(repo, {}).get(cp)
                if cpdict:
                    pkg_useforce = ordered_by_atom_specificity(cpdict, pkg)
                    if pkg_useforce:
                        useforce.extend(pkg_useforce)
                if stable:
                    cpdict = self._repo_pusestableforce_dict.get(repo, {}).get(cp)
                    if cpdict:
                        pkg_useforce = ordered_by_atom_specificity(cpdict, pkg)
                        if pkg_useforce:
                            useforce.extend(pkg_useforce)

        for i, puseforce_dict in enumerate(self._puseforce_list):
            if self._useforce_list[i]:
                useforce.append(self._useforce_list[i])
            if stable and self._usestableforce_list[i]:
                useforce.append(self._usestableforce_list[i])
            cpdict = puseforce_dict.get(cp)
            if cpdict:
                pkg_useforce = ordered_by_atom_specificity(cpdict, pkg)
                if pkg_useforce:
                    useforce.extend(pkg_useforce)
            if stable:
                cpdict = self._pusestableforce_list[i].get(cp)
                if cpdict:
                    pkg_useforce = ordered_by_atom_specificity(cpdict, pkg)
                    if pkg_useforce:
                        useforce.extend(pkg_useforce)

        return frozenset(stack_lists(useforce, incremental=True))

    def getPUSE(self, pkg):
        cp = getattr(pkg, "cp", None)
        if cp is None:
            slot = dep_getslot(pkg)
            repo = dep_getrepo(pkg)
            pkg = _pkg_str(remove_slot(pkg), slot=slot, repo=repo)
            cp = pkg.cp
        ret = ""
        cpdict = self._pusedict.get(cp)
        if cpdict:
            puse_matches = ordered_by_atom_specificity(cpdict, pkg)
            if puse_matches:
                puse_list = []
                for x in puse_matches:
                    puse_list.extend(x)
                ret = " ".join(puse_list)
        return ret

    def extract_global_USE_changes(self, old=""):
        ret = old
        cpdict = self._pusedict.get("*/*")
        if cpdict is not None:
            v = cpdict.pop("*/*", None)
            if v is not None:
                ret = " ".join(v)
                if old:
                    ret = old + " " + ret
                if not cpdict:
                    # No tokens left in atom_license_map, remove it.
                    del self._pusedict["*/*"]
        return ret



# Copyright 1999-2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import errno
import re
import stat
import sys
import warnings

from portage import os
from portage import _encodings
from portage import _unicode_decode
from portage import _unicode_encode

from portage.const import USER_CONFIG_PATH, VCS_DIRS
from portage.eapi import _get_eapi_attrs
from portage.exception import DirectoryNotFound, InvalidAtom, PortageException
from portage.localization import _


ignored_dbentries = ("CONTENTS", "environment.bz2")


def update_dbentry(update_cmd, mycontent, eapi=None, parent=None):
    from portage.dep import Atom, isvalidatom, match_from_list

    if parent is not None:
        eapi = parent.eapi

    if update_cmd[0] == "move":
        old_value = str(update_cmd[1])
        new_value = str(update_cmd[2])

        # Use isvalidatom() to check if this move is valid for the
        # EAPI (characters allowed in package names may vary).
        if old_value in mycontent and isvalidatom(new_value, eapi=eapi):
            # this split preserves existing whitespace
            split_content = re.split(r"(\s+)", mycontent)
            modified = False
            for i, token in enumerate(split_content):
                if old_value not in token:
                    continue
                try:
                    atom = Atom(token, eapi=eapi)
                except InvalidAtom:
                    continue
                if atom.cp != old_value:
                    continue

                new_atom = Atom(token.replace(old_value, new_value, 1), eapi=eapi)

                # Avoid creating self-blockers for bug #367215.
                if (
                    new_atom.blocker
                    and parent is not None
                    and parent.cp == new_atom.cp
                    and match_from_list(new_atom, [parent])
                ):
                    continue

                split_content[i] = str(new_atom)
                modified = True

            if modified:
                mycontent = "".join(split_content)

    elif update_cmd[0] == "slotmove" and update_cmd[1].operator is None:
        orig_atom, origslot, newslot = update_cmd[1:]
        orig_cp = orig_atom.cp

        # We don't support versioned slotmove atoms here, since it can be
        # difficult to determine if the version constraints really match
        # the atoms that we're trying to update.
        if orig_atom.version is None and orig_cp in mycontent:
            # this split preserves existing whitespace
            split_content = re.split(r"(\s+)", mycontent)
            modified = False
            for i, token in enumerate(split_content):
                if orig_cp not in token:
                    continue
                try:
                    atom = Atom(token, eapi=eapi)
                except InvalidAtom:
                    continue
                if atom.cp != orig_cp:
                    continue
                if atom.slot is None or atom.slot != origslot:
                    continue

                slot_part = newslot
                if atom.sub_slot is not None:
                    if atom.sub_slot == origslot:
                        sub_slot = newslot
                    else:
                        sub_slot = atom.sub_slot
                    slot_part += "/" + sub_slot
                if atom.slot_operator is not None:
                    slot_part += atom.slot_operator

                split_content[i] = atom.with_slot(slot_part)
                modified = True

            if modified:
                mycontent = "".join(split_content)

    return mycontent


def update_dbentries(update_iter, mydata, eapi=None, parent=None):
    """Performs update commands and returns a
    dict containing only the updated items."""
    updated_items = {}
    for k, mycontent in mydata.items():
        k_unicode = _unicode_decode(
            k, encoding=_encodings["repo.content"], errors="replace"
        )
        if k_unicode not in ignored_dbentries:
            orig_content = mycontent
            mycontent = _unicode_decode(
                mycontent, encoding=_encodings["repo.content"], errors="replace"
            )
            is_encoded = mycontent is not orig_content
            orig_content = mycontent
            for update_cmd in update_iter:
                mycontent = update_dbentry(
                    update_cmd, mycontent, eapi=eapi, parent=parent
                )
            if mycontent != orig_content:
                if is_encoded:
                    mycontent = _unicode_encode(
                        mycontent,
                        encoding=_encodings["repo.content"],
                        errors="backslashreplace",
                    )
                updated_items[k] = mycontent
    return updated_items


def fixdbentries(update_iter, dbdir, eapi=None, parent=None):
    """Performs update commands which result in search and replace operations
    for each of the files in dbdir (excluding CONTENTS and environment.bz2).
    Returns True when actual modifications are necessary and False otherwise."""
    from portage.util import write_atomic

    warnings.warn(
        "portage.update.fixdbentries() is deprecated", DeprecationWarning, stacklevel=2
    )

    mydata = {}
    for myfile in [f for f in os.listdir(dbdir) if f not in ignored_dbentries]:
        file_path = os.path.join(dbdir, myfile)
        with open(
            _unicode_encode(file_path, encoding=_encodings["fs"], errors="strict"),
            encoding=_encodings["repo.content"],
            errors="replace",
        ) as f:
            mydata[myfile] = f.read()
    updated_items = update_dbentries(update_iter, mydata, eapi=eapi, parent=parent)
    for myfile, mycontent in updated_items.items():
        file_path = os.path.join(dbdir, myfile)
        write_atomic(file_path, mycontent, encoding=_encodings["repo.content"])
    return len(updated_items) > 0


def grab_updates(updpath, prev_mtimes=None):
    """Returns all the updates from the given directory as a sorted list of
    tuples, each containing (file_path, statobj, content).  If prev_mtimes is
    given then updates are only returned if one or more files have different
    mtimes. When a change is detected for a given file, updates will be
    returned for that file and any files that come after it in the entire
    sequence. This ensures that all relevant updates are returned for cases
    in which the destination package of an earlier move corresponds to
    the source package of a move that comes somewhere later in the entire
    sequence of files.
    """
    try:
        mylist = os.listdir(updpath)
    except OSError as oe:
        if oe.errno == errno.ENOENT:
            raise DirectoryNotFound(updpath)
        raise
    if prev_mtimes is None:
        prev_mtimes = {}

    update_data = []
    for myfile in mylist:
        if myfile.startswith("."):
            continue
        file_path = os.path.join(updpath, myfile)
        mystat = os.stat(file_path)
        if not stat.S_ISREG(mystat.st_mode):
            continue
        if int(prev_mtimes.get(file_path, -1)) != mystat[stat.ST_MTIME]:
            f = open(
                _unicode_encode(file_path, encoding=_encodings["fs"], errors="strict"),
                encoding=_encodings["repo.content"],
                errors="replace",
            )
            content = f.read()
            f.close()
            update_data.append((file_path, mystat, content))
    return update_data


def parse_updates(mycontent):
    """Valid updates are returned as a list of split update commands."""
    from portage.dep import Atom
    from portage.versions import _get_slot_re

    eapi_attrs = _get_eapi_attrs(None)
    slot_re = _get_slot_re(eapi_attrs)
    myupd = []
    errors = []
    mylines = mycontent.splitlines()
    for myline in mylines:
        mysplit = myline.split()
        if len(mysplit) == 0:
            continue
        if mysplit[0] not in ("move", "slotmove"):
            errors.append(_("ERROR: Update type not recognized '%s'") % myline)
            continue
        if mysplit[0] == "move":
            if len(mysplit) != 3:
                errors.append(_("ERROR: Update command invalid '%s'") % myline)
                continue
            valid = True
            for i in (1, 2):
                try:
                    atom = Atom(mysplit[i])
                except InvalidAtom:
                    atom = None
                else:
                    if atom.blocker or atom != atom.cp:
                        atom = None
                if atom is not None:
                    mysplit[i] = atom
                else:
                    errors.append(_("ERROR: Malformed update entry '%s'") % myline)
                    valid = False
                    break
            if not valid:
                continue

        if mysplit[0] == "slotmove":
            if len(mysplit) != 4:
                errors.append(_("ERROR: Update command invalid '%s'") % myline)
                continue
            pkg, origslot, newslot = mysplit[1], mysplit[2], mysplit[3]
            try:
                atom = Atom(pkg)
            except InvalidAtom:
                atom = None
            else:
                if atom.blocker:
                    atom = None
            if atom is not None:
                mysplit[1] = atom
            else:
                errors.append(_("ERROR: Malformed update entry '%s'") % myline)
                continue

            invalid_slot = False
            for slot in (origslot, newslot):
                m = slot_re.match(slot)
                if m is None:
                    invalid_slot = True
                    break
                if "/" in slot:
                    # EAPI 5 style subslot is currently not supported.
                    invalid_slot = True
                    break

            if invalid_slot:
                errors.append(_("ERROR: Malformed update entry '%s'") % myline)
                continue

        # The list of valid updates is filtered by continue statements above.
        myupd.append(mysplit)
    return myupd, errors


def update_config_files(
    config_root,
    protect,
    protect_mask,
    update_iter,
    match_callback=None,
    case_insensitive=False,
):
    """Perform global updates on /etc/portage/package.*, /etc/portage/profile/package.*,
    /etc/portage/profile/packages and /etc/portage/sets.
    config_root - location of files to update
    protect - list of paths from CONFIG_PROTECT
    protect_mask - list of paths from CONFIG_PROTECT_MASK
    update_iter - list of update commands as returned from parse_updates(),
            or dict of {repo_name: list}
    match_callback - a callback which will be called with three arguments:
            match_callback(repo_name, old_atom, new_atom)
    and should return boolean value determining whether to perform the update"""
    from portage.dep import isvalidatom
    from portage.util import (
        ConfigProtect,
        new_protect_filename,
        normalize_path,
        write_atomic,
        writemsg,
    )

    repo_dict = None
    if isinstance(update_iter, dict):
        repo_dict = update_iter
    if match_callback is None:

        def match_callback(repo_name, atoma, atomb):
            return True

    config_root = normalize_path(config_root)
    update_files = {}
    file_contents = {}
    myxfiles = [
        "package.accept_keywords",
        "package.env",
        "package.keywords",
        "package.license",
        "package.mask",
        "package.properties",
        "package.unmask",
        "package.use",
        "sets",
    ]
    myxfiles += [
        os.path.join("profile", x)
        for x in (
            "packages",
            "package.accept_keywords",
            "package.keywords",
            "package.mask",
            "package.unmask",
            "package.use",
            "package.use.force",
            "package.use.mask",
            "package.use.stable.force",
            "package.use.stable.mask",
        )
    ]
    abs_user_config = os.path.join(config_root, USER_CONFIG_PATH)
    recursivefiles = []
    for x in myxfiles:
        config_file = os.path.join(abs_user_config, x)
        if os.path.isdir(config_file):
            for parent, dirs, files in os.walk(config_file):
                try:
                    parent = _unicode_decode(
                        parent, encoding=_encodings["fs"], errors="strict"
                    )
                except UnicodeDecodeError:
                    continue
                for y_enc in list(dirs):
                    try:
                        y = _unicode_decode(
                            y_enc, encoding=_encodings["fs"], errors="strict"
                        )
                    except UnicodeDecodeError:
                        dirs.remove(y_enc)
                        continue
                    if y.startswith(".") or y in VCS_DIRS:
                        dirs.remove(y_enc)
                for y in files:
                    try:
                        y = _unicode_decode(
                            y, encoding=_encodings["fs"], errors="strict"
                        )
                    except UnicodeDecodeError:
                        continue
                    if y.startswith("."):
                        continue
                    recursivefiles.append(
                        os.path.join(parent, y)[len(abs_user_config) + 1 :]
                    )
        else:
            recursivefiles.append(x)
    myxfiles = recursivefiles
    for x in myxfiles:
        f = None
        try:
            f = open(
                _unicode_encode(
                    os.path.join(abs_user_config, x),
                    encoding=_encodings["fs"],
                    errors="strict",
                ),
                encoding=_encodings["content"],
                errors="replace",
            )
            file_contents[x] = f.readlines()
        except OSError:
            continue
        finally:
            if f is not None:
                f.close()

    ignore_line_re = re.compile(r"^#|^\s*$")
    if repo_dict is None:
        update_items = [(None, update_iter)]
    else:
        update_items = [x for x in repo_dict.items() if x[0] != "DEFAULT"]
    for repo_name, update_iter in update_items:
        for update_cmd in update_iter:
            for x, contents in file_contents.items():
                skip_next = False
                for pos, line in enumerate(contents):
                    if skip_next:
                        skip_next = False
                        continue
                    if ignore_line_re.match(line):
                        continue
                    atom = line.split()[0]
                    if atom[:1] == "-":
                        # package.mask supports incrementals
                        atom = atom[1:]
                    if atom[:1] == "*":
                        # packages file supports "*"-prefixed atoms as indication of system packages.
                        atom = atom[1:]
                    if not isvalidatom(atom):
                        continue
                    new_atom = update_dbentry(update_cmd, atom)
                    if atom != new_atom:
                        if match_callback(repo_name, atom, new_atom):
                            # add a comment with the update command, so
                            # the user can clearly see what happened
                            contents[pos] = "# {}\n".format(
                                " ".join(f"{x}" for x in update_cmd)
                            )
                            contents.insert(
                                pos + 1,
                                line.replace(f"{atom}", f"{new_atom}", 1),
                            )
                            # we've inserted an additional line, so we need to
                            # skip it when it's reached in the next iteration
                            skip_next = True
                            update_files[x] = 1
                            sys.stdout.write("p")
                            sys.stdout.flush()

    protect_obj = ConfigProtect(
        config_root, protect, protect_mask, case_insensitive=case_insensitive
    )
    for x in update_files:
        updating_file = os.path.join(abs_user_config, x)
        if protect_obj.isprotected(updating_file):
            updating_file = new_protect_filename(updating_file)
        try:
            write_atomic(updating_file, "".join(file_contents[x]))
        except PortageException as e:
            writemsg(f"\n!!! {str(e)}\n", noiselevel=-1)
            writemsg(
                _("!!! An error occurred while updating a config file:")
                + f" '{updating_file}'\n",
                noiselevel=-1,
            )
            continue


def dep_transform(mydep, oldkey, newkey):
    from portage.dep import dep_getkey

    if dep_getkey(mydep) == oldkey:
        return mydep.replace(oldkey, newkey, 1)
    return mydep
